# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Console script for broadlink_listener."""

import inspect
import logging
import sys
import types
from pathlib import Path
from typing import cast

import click
from broadlink import discover
from broadlink.const import DEFAULT_BCAST_ADDR, DEFAULT_TIMEOUT
from cloup import HelpFormatter, HelpTheme, Style, group, option, option_group

from broadlink_listener import __version__
from broadlink_listener.cli_tools.broadlink_manager import BroadlinkManager
from broadlink_listener.cli_tools.smartir_manager import SmartIrManager
from broadlink_listener.cli_tools.utils import configure_logger, get_local_ip_address

formatter_settings = HelpFormatter.settings(
    theme=HelpTheme(
        invoked_command=Style(fg='bright_yellow'),
        heading=Style(fg='bright_white', bold=True),
        constraint=Style(fg='magenta'),
        col1=Style(fg='bright_yellow'),
    )
)


@group(align_option_groups=False, formatter_settings=formatter_settings, no_args_is_help=False)  # type: ignore
@option_group(
    "Generic Options",
    option(
        '-d',
        '--loglevel',
        'loglevel',
        type=click.Choice(["debug", "info", "warning", "error"], case_sensitive=False),
        default="info",
        show_default=True,
        help="set log level",
    ),
)
@click.version_option(__version__)
def main(loglevel: str):
    """Broadlink listener.

    # noqa: DAR101
    """
    _fnc_name = cast(types.FrameType, inspect.currentframe()).f_code.co_name
    _complete_doc = inspect.getdoc(eval(_fnc_name))  # pylint: disable=eval-used  # nosec B307
    _doc = f"{_complete_doc}".split('\n')[0]  # use only doc first row
    _str = f"{_doc[:-1]} - v{__version__}"
    click.echo(f"{_str}")
    click.echo("=" * len(_str))
    click.echo("Broadlink IR codes listener and SmartIR json generator.")

    # ctx.ensure_object(dict)
    # ctx.obj = loglevel
    configure_logger(loglevel)


@main.command(help="Discover Broadlink IR")
@click.argument('local_ip', type=str, default=get_local_ip_address(), required=True)
def discover_ir(local_ip: str):
    """Discover Broadlink IR, code from python-broadlink.

    Arguments:
        local_ip: IP address of this machine connected to same network Broadlink IR is connected to.
    """
    devices = discover(timeout=DEFAULT_TIMEOUT, local_ip_address=local_ip, discover_ip_address=DEFAULT_BCAST_ADDR)
    for device in devices:
        if device.auth():
            click.echo("###########################################")
            click.echo(f"Device Type: {device.type}")
            click.echo(
                f"Broadlink Type: {hex(device.devtype)}\n"
                f"Broadlink IP Address: {device.host[0]} "
                f"Broadlink MAC Address: {''.join(format(x, '02x') for x in device.mac)}"
            )
        else:
            logging.error("Error authenticating with device : %s", device.host)


@main.command(help="Generate SmartIR json file from input one")
@click.argument('json_file', type=click.Path(exists=True), required=True)
@click.argument('dev_type', type=str, required=True)
@click.argument('ip_addr', type=str, required=True)
@click.argument('mac_addr', type=str, required=True)
@click.option('--no-temp-on-mode', '-n', type=str, multiple=True, default=[])
@click.option('--no-swing-on-mode', '-s', type=str, multiple=True, default=[])
def generate_smart_ir(
    json_file: Path, dev_type: str, ip_addr: str, mac_addr: str, no_temp_on_mode: tuple, no_swing_on_mode: tuple
):  # pylint: disable=too-many-arguments, too-many-locals
    """Generate SmartIR json file from input one.

    Arguments:
        json_file: SmartIR json file that contains basic structure of codes to be recorded.
        dev_type: Broadlink device type from discover IR command
        ip_addr: Broadlink IP address from discover IR command
        mac_addr: Broadlink MAC address from discover IR command
        no_temp_on_mode: option, that can be set multiple times, related to operating mode that have no temperature
                         selection
        no_swing_on_mode: option, that can be set multiple times, related to operating mode that have no swing
                          selection
    """
    broadlink_mng = BroadlinkManager(dev_type, ip_addr, mac_addr)
    smart_ir_mng = SmartIrManager(json_file, broadlink_mng, no_temp_on_mode, no_swing_on_mode)

    smart_ir_mng.learn_off()
    smart_ir_mng.learn_all()
    smart_ir_mng.save_dict()


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover  # pylint: disable=no-value-for-parameter

    # local test
    # sys.exit(main(['discover-ir']))



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""SmartIR json manager class."""

import glob
import json
import os
import platform
import re
import signal
import sys
from collections import namedtuple
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from itertools import product
from pathlib import Path
from threading import Event
from typing import Literal, Optional, Union

import click

from broadlink_listener.cli_tools.broadlink_manager import BroadlinkManager


class _DictKeys(str, Enum):
    CONTROLLER = "supportedController"
    TEMPERATURE = "temperature"
    MIN_TEMP = "minTemperature"
    MAX_TEMP = "maxTemperature"
    PRECISION = "precision"
    OPERATION_MODES = "operationModes"
    FAN_MODES = "fanModes"
    SWING_MODES = "swingModes"
    COMMANDS = "commands"
    COMMANDS_ENCODING = "commandsEncoding"


@dataclass
class SkipReason:
    """Data class to describe skip reason."""

    skip: bool
    field: Optional[Union[Literal[_DictKeys.SWING_MODES], Literal[_DictKeys.TEMPERATURE]]]


_combination_arguments_all = (
    _DictKeys.OPERATION_MODES.value,
    _DictKeys.FAN_MODES.value,
    _DictKeys.SWING_MODES.value,
    _DictKeys.TEMPERATURE.value,
)

_combination_arguments_swing = (
    _DictKeys.OPERATION_MODES.value,
    _DictKeys.SWING_MODES.value,
    _DictKeys.TEMPERATURE.value,
)

_combination_arguments_fan = (
    _DictKeys.OPERATION_MODES.value,
    _DictKeys.FAN_MODES.value,
    _DictKeys.TEMPERATURE.value,
)

_combination_arguments_none = (
    _DictKeys.OPERATION_MODES.value,
    _DictKeys.TEMPERATURE.value,
)


_CombinationTupleAll = namedtuple('_CombinationTupleAll', ', '.join(_combination_arguments_all))  # type: ignore

_CombinationTupleSwing = namedtuple('_CombinationTupleSwing', ', '.join(_combination_arguments_swing))  # type: ignore

_CombinationTupleFan = namedtuple('_CombinationTupleFan', ', '.join(_combination_arguments_fan))  # type: ignore

_CombinationTupleNone = namedtuple('_CombinationTupleNone', ', '.join(_combination_arguments_none))  # type: ignore


def _countdown(msg: str, event: Event):
    click.echo(msg)
    if event.is_set():
        input("-->> Press enter when ready <<--")
    event.clear()


class SmartIrManager:  # pylint: disable=too-many-instance-attributes
    """Manager class for SmartIR json."""

    def __init__(  # pylint: disable=too-many-branches,too-many-statements
        self,
        file_name: Path,
        broadlink_mng: BroadlinkManager,
        no_temp_on_mode: tuple = tuple(),
        no_swing_on_mode: tuple = tuple(),
    ):
        """Smart IR Manager.

        Arguments:
            file_name: SmartIR json file that contains basic structure of codes to be recorded.
            broadlink_mng: Broadlink Manager object used to listen IR codes
            no_temp_on_mode: option, that can be set multiple times, related to operating mode that have no temperature
                         selection
            no_swing_on_mode: option, that can be set multiple times, related to operating mode that have no swing
                          selection

        Raises:
            UsageError: raised if controller is not Broadlink or no IR signal is learnt during the process
        """
        self.__broadlink_manager = broadlink_mng

        self.__partial_inc = 0
        self.__prompt_event = Event()
        self.__prompt_event.clear()
        self.__json_file_name_path = Path(file_name)
        with open(str(file_name), "r", encoding='utf-8') as in_file:
            self.__smartir_dict = json.load(in_file)

        self.__all_combinations: tuple = ()
        self.__no_temp_on_modes: tuple = no_temp_on_mode
        self.__no_swing_on_modes: tuple = no_swing_on_mode
        self.__swing_saved_temp: dict = {}
        try:
            _controller = self.__smartir_dict[_DictKeys.CONTROLLER.value]
            if _controller != "Broadlink":
                raise click.exceptions.UsageError(f"Controller {_controller} not supported")

            _commands_encoding = self.__smartir_dict[_DictKeys.COMMANDS_ENCODING.value]
            if _commands_encoding != "Base64":
                raise click.exceptions.UsageError(f"Encoding {_commands_encoding} not supported")

            self.__min_temp = int(self.__smartir_dict[_DictKeys.MIN_TEMP.value])
            self.__max_temp = int(self.__smartir_dict[_DictKeys.MAX_TEMP.value])
            self.__precision_temp = int(self.__smartir_dict.get(_DictKeys.PRECISION.value, 1))
            self.__op_modes = self.__smartir_dict[_DictKeys.OPERATION_MODES.value]
            self.__fan_modes = self.__smartir_dict.get(_DictKeys.FAN_MODES.value, None)
            self.__swing_modes = self.__smartir_dict.get(_DictKeys.SWING_MODES.value, None)

            if not all(list(map(lambda x: x in self.__op_modes, no_swing_on_mode))):
                raise click.exceptions.UsageError("no-swing-on-mode parameter is using a not-existent operating mode.")

            if not all(list(map(lambda x: x in self.__op_modes, no_temp_on_mode))):
                raise click.exceptions.UsageError("no-temp-on-mode parameter is using a not-existent operating mode.")

            # fill dict with all empty combination
            _temp_dict = {
                f"{t}": deepcopy('') for t in range(self.__min_temp, self.__max_temp + 1, self.__precision_temp)
            }
            _swing_dict = None
            if self.__swing_modes:
                _swing_dict = {f"{s}": deepcopy(_temp_dict) for s in self.__swing_modes}
            else:
                if no_swing_on_mode:
                    raise click.exceptions.UsageError(
                        "Add swing to JSON file or do not use --no-swing-on-mode parameter."
                    )
            _fan_mode_dict = None
            if self.__fan_modes:
                if _swing_dict:
                    _fan_mode_dict = {f"{f}": deepcopy(_swing_dict) for f in self.__fan_modes}
                else:
                    _fan_mode_dict = {f"{f}": deepcopy(_temp_dict) for f in self.__fan_modes}  # type: ignore
                _operation_dict = {f"{o}": deepcopy(_fan_mode_dict) for o in self.__op_modes}
            else:
                if _swing_dict:
                    _operation_dict = {f"{o}": deepcopy(_swing_dict) for o in self.__op_modes}  # type: ignore
                else:
                    _operation_dict = {f"{o}": deepcopy(_temp_dict) for o in self.__op_modes}  # type: ignore
            self.__smartir_dict[_DictKeys.COMMANDS].update(_operation_dict)

            # overwrite combination if tmp file exist
            self._load_partial_dict()

        except KeyError as key_err:
            raise click.exceptions.UsageError(f"Missing mandatory field in json file: {key_err}") from None

        self._setup_combinations()
        self.__temperature = ''
        self.__operation_mode = ''
        self.__fan_mode = ''
        self.__swing_mode = ''
        self._setup_signal_handler()

    @property
    def smartir_dict(self) -> dict:
        """Output dictionary, for test purpose.

        Returns:
            dict: smartir compatible dictionary with learnt codes.
        """
        return self.__smartir_dict

    def _setup_signal_handler(self):
        _system = platform.system().lower()
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGABRT, self._signal_handler)

        if _system in ('linux', 'darwin'):
            signal.signal(signal.SIGTERM, self._signal_handler)
        if _system == 'windows':
            signal.signal(signal.SIGBREAK, self._signal_handler)  # pylint: disable=no-member

    def _signal_handler(self, _signumber, _frame):
        self._save_partial_dict()
        sys.exit(2)

    def _setup_combinations(self):
        _variable_args = [self.__fan_modes, self.__swing_modes]
        if all(_variable_args):
            __combinations = product(
                self.__op_modes,
                self.__fan_modes,
                self.__swing_modes,
                range(self.__min_temp, self.__max_temp + 1, self.__precision_temp),
            )

            def _return_named_tuple():
                for _c in __combinations:
                    yield _CombinationTupleAll(_c[0], _c[1], _c[2], _c[3])

            self.__all_combinations = _return_named_tuple()
            self.__combination_arguments = _combination_arguments_all
        else:
            if any(_variable_args):
                if self.__swing_modes:
                    __combinations = product(
                        self.__op_modes,
                        self.__swing_modes,
                        range(self.__min_temp, self.__max_temp + 1, self.__precision_temp),
                    )

                    def _return_named_tuple():
                        for _c in __combinations:
                            yield _CombinationTupleSwing(_c[0], _c[1], _c[2])

                    self.__all_combinations = _return_named_tuple()
                    self.__combination_arguments = _combination_arguments_swing
                else:
                    __combinations = product(
                        self.__op_modes,
                        self.__fan_modes,
                        range(self.__min_temp, self.__max_temp + 1, self.__precision_temp),
                    )

                    def _return_named_tuple():
                        for _c in __combinations:
                            yield _CombinationTupleFan(_c[0], _c[1], _c[2])

                    self.__all_combinations = _return_named_tuple()
                    self.__combination_arguments = _combination_arguments_fan
            else:
                __combinations = product(
                    self.__op_modes, range(self.__min_temp, self.__max_temp + 1, self.__precision_temp)
                )

                def _return_named_tuple():
                    for _c in __combinations:
                        yield _CombinationTupleNone(_c[0], _c[1])

                self.__all_combinations = _return_named_tuple()
                self.__combination_arguments = _combination_arguments_none

    @property
    def temperature(self) -> str:
        """Temperature key saved to json.

        Returns:
            str: temperature string to be saved
        """
        return self.__temperature

    @temperature.setter
    def temperature(self, new_value: str) -> None:
        """Set Temperature key.

        Arguments:
            new_value: value to be set
        """
        self.__temperature = new_value

    @property
    def operation_mode(self) -> str:
        """Operation Mode key saved to json.

        Returns:
            str: operation mode string to be saved
        """
        return self.__operation_mode

    @operation_mode.setter
    def operation_mode(self, new_value: str) -> None:
        """Set Operation Mode key.

        Arguments:
            new_value: value to be set
        """
        self.__operation_mode = new_value

    @property
    def fan_mode(self) -> str:
        """Fan Mode key saved to json.

        Returns:
            str: fan mode string to be saved
        """
        return self.__fan_mode

    @fan_mode.setter
    def fan_mode(self, new_value: str) -> None:
        """Set Fan Mode key.

        Arguments:
            new_value: value to be set
        """
        self.__fan_mode = new_value

    @property
    def swing_mode(self) -> str:
        """Swing Mode key saved to json.

        Returns:
            str: swing mode string to be saved
        """
        return self.__swing_mode

    @swing_mode.setter
    def swing_mode(self, new_value: str) -> None:
        """Set Swing Mode key.

        Arguments:
            new_value: value to be set
        """
        self.__swing_mode = new_value

    @property
    def partial_inc(self) -> int:
        """Partial increment value.

        Returns:
            int: last index of saved partial json files
        """
        return self.__partial_inc

    def _set_dict_value(self, value: str) -> None:
        if _DictKeys.FAN_MODES in self.__combination_arguments:
            if _DictKeys.SWING_MODES in self.__combination_arguments:
                self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.fan_mode][self.swing_mode][
                    self.temperature
                ] = value
            else:
                self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.fan_mode][
                    self.temperature
                ] = value
        else:
            if _DictKeys.SWING_MODES in self.__combination_arguments:
                self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.swing_mode][
                    self.temperature
                ] = value
            else:
                self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.temperature] = value

    def _get_dict_value(self) -> str:
        _value = ''
        if _DictKeys.FAN_MODES in self.__combination_arguments:
            if _DictKeys.SWING_MODES in self.__combination_arguments:
                _value = self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.fan_mode][
                    self.swing_mode
                ][self.temperature]
            else:
                _value = self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.fan_mode][
                    self.temperature
                ]
        else:
            if _DictKeys.SWING_MODES in self.__combination_arguments:
                _value = self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.swing_mode][
                    self.temperature
                ]
            else:
                _value = self.__smartir_dict[_DictKeys.COMMANDS.value][self.operation_mode][self.temperature]
        return _value

    def save_dict(self):
        """Save modified dict to output json file."""
        now = datetime.now()
        _modified_file_name = self.__json_file_name_path.parent.joinpath(
            f'{self.__json_file_name_path.stem}_{now.strftime("%Y%m%d_%H%M%S")}.json'
        )
        with open(_modified_file_name, 'w', encoding='utf-8') as out_file:
            json.dump(self.__smartir_dict, out_file)
        click.echo(f"Created new file {_modified_file_name}")
        _previous = glob.glob(
            os.path.join(self.__json_file_name_path.parent, f'{self.__json_file_name_path.stem}_tmp_*.json')
        )
        for _file in _previous:
            os.remove(_file)

    def _save_partial_dict(self):
        # save with incremental 3 numbers to sort correctly when load
        _modified_file_name = self.__json_file_name_path.parent.joinpath(
            f'{self.__json_file_name_path.stem}_tmp_{self.__partial_inc:03}.json'
        )

        try:
            _no_off = deepcopy(self.__smartir_dict)
            del _no_off['commands']['off']
        except KeyError:
            return

        with open(_modified_file_name, 'w', encoding='utf-8') as out_file:
            json.dump(_no_off[_DictKeys.COMMANDS.value], out_file, indent=2)
        self.__partial_inc += 1

    def _load_partial_dict(self):
        _previous = glob.glob(
            os.path.join(self.__json_file_name_path.parent, f'{self.__json_file_name_path.stem}_tmp_*.json')
        )
        _previous.sort()
        try:
            # load last file that's the most updated
            _last_file = _previous[-1]
            with open(str(_last_file), "r", encoding='utf-8') as partial_file:
                self.__smartir_dict[_DictKeys.COMMANDS.value].update(json.load(partial_file))
            _pattern = re.compile(fr'{self.__json_file_name_path.stem}_tmp_(\d+).json')
            _res = _pattern.search(_last_file)
            self.__partial_inc = int(_res.group(1))
        except IndexError:
            pass

    def learn_off(self):
        """Learn OFF command that's outside the combination.

        Raises:
            UsageError: if no IR signal is learnt within timeout
        """
        self.__prompt_event.clear()
        _countdown(
            "First of all, let's learn OFF command:\nturn ON the remote and then turn it OFF when "
            "'Listening' message is on screen, or interrupt with CTRL-C...",
            self.__prompt_event,
        )
        # set event to wait for first code
        self.__prompt_event.set()
        _off = self.__broadlink_manager.learn_single_code()
        if not _off:
            raise click.exceptions.UsageError("No IR signal learnt for OFF command within timeout.")
        self.__smartir_dict[_DictKeys.COMMANDS.value]["off"] = _off

    def learn_all(self):  # pylint: disable=too-many-branches
        """Learn all the commands depending on calculated combination.

        Raises:
            UsageError: if no IR signal is learnt within timeout
        """
        _previous_code = None
        _previous_combination: Optional[tuple] = None
        for comb in self.__all_combinations:  # pylint: disable=too-many-nested-blocks
            self.operation_mode = comb.operationModes
            if _DictKeys.FAN_MODES in comb._fields:
                self.fan_mode = comb.fanModes
            if _DictKeys.SWING_MODES in comb._fields:
                self.swing_mode = comb.swingModes
            self.temperature = str(comb.temperature)

            if self._get_dict_value() != '':
                self.__prompt_event.set()
                continue

            _do_skip = self._skip_learning(comb)
            if _do_skip.skip:
                # must read the first temperature and then reuse the same for next combination
                if _do_skip.field == _DictKeys.TEMPERATURE:
                    if comb.temperature > self.__min_temp:
                        # code @ min_temp already recorded
                        self._set_dict_value(_previous_code)
                        continue

                if _do_skip.field == _DictKeys.SWING_MODES:
                    _previous_code = self.__swing_saved_temp.get(self.temperature, None)
                    if _previous_code:
                        if isinstance(_previous_code, dict):
                            # there also is the fan level saved
                            _temp_code = _previous_code.get(self.fan_mode)
                            if _temp_code:
                                self._set_dict_value(_temp_code)
                                continue
                        else:
                            # no fan
                            self._set_dict_value(_previous_code)
                            continue

            if _previous_combination:
                for i in range(0, len(comb) - 1):
                    if _previous_combination[i] != comb[i]:  # pylint: disable=unsubscriptable-object
                        self.__prompt_event.set()
                        self._save_partial_dict()
            _previous_combination = comb

            _combination_str = self._get_combination(comb)
            _countdown(
                "-" * 30 + f"\nLet's learn IR command of\n{_combination_str}\n"
                "Prepare the remote so Broadlink can listen the above combination when 'Listening' message"
                " is on screen, or interrupt with CTRL-C...",
                self.__prompt_event,
            )
            _code = self.__broadlink_manager.learn_single_code()
            _previous_code = _code
            if not _code:
                self._save_partial_dict()
                raise click.exceptions.UsageError(f"No IR signal learnt for {_combination_str} command within timeout.")

            # swing modes must be saved because all temperature need to be listened
            if _do_skip.skip and _do_skip.field == _DictKeys.SWING_MODES:
                if _DictKeys.FAN_MODES in comb._fields:
                    self.__swing_saved_temp[self.temperature] = {self.fan_mode: _code}
                else:
                    self.__swing_saved_temp[self.temperature] = _code

            self._set_dict_value(_code)
        click.echo("All combination learnt.")

    def _get_combination(self, combination: tuple) -> str:
        _mixed = zip(self.__combination_arguments, combination)
        _ret = []
        for _m, _v in _mixed:
            _ret.append(f'-> {_m} = {_v}')
        return '\n'.join(_ret)

    def _skip_learning(
        self, comb: Union[_CombinationTupleAll, _CombinationTupleSwing, _CombinationTupleFan, _CombinationTupleNone]
    ) -> SkipReason:
        _ret = SkipReason(False, None)
        # swing mode is optional, so need its check
        if _DictKeys.SWING_MODES in comb._fields and comb.operationModes in self.__no_swing_on_modes:  # type: ignore
            _ret = SkipReason(True, _DictKeys.SWING_MODES)
        if comb.operationModes in self.__no_temp_on_modes:  # type: ignore
            _ret = SkipReason(True, _DictKeys.TEMPERATURE)
        return _ret



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Pytest conftest."""

import binascii
import glob
import json
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Generator

import pytest
from click.testing import CliRunner


def _remove_tmp_files(file_pattern=r"[a-zA-Z_]*_tmp[_\d]+.json"):
    """Utility method to remove test generated files.

    Arguments:
        file_pattern: regex pattern for files to be removed.
    """
    _pattern = re.compile(file_pattern)
    _previous = glob.glob(os.path.join(Path.cwd().joinpath("tests").joinpath("data"), '*'))
    for _file in _previous:
        if _pattern.match(os.path.basename(_file)):
            os.remove(_file)


@pytest.fixture
def json_file_not_broadlink() -> Path:
    """Return json test file with not supported controller.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("data").joinpath("not_broadlink.json")


@pytest.fixture
def json_file_not_base64() -> Path:
    """Return json test file with not supported encoding.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("data").joinpath("not_base64.json")


@pytest.fixture
def json_file_missing_max_temp() -> Path:
    """Return json test file without max temp.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("data").joinpath("missing_required_max_temp.json")


@pytest.fixture
def json_file_missing_min_temp() -> Path:
    """Return json test file without min temp.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("data").joinpath("missing_required_min_temp.json")


@pytest.fixture
def json_file_good_data_op_mode() -> Generator:
    """Return json test file with good structure.

    Yields:
        json's path file
    """
    yield Path.cwd().joinpath("tests").joinpath("data").joinpath("good_data_op_mode.json")
    _remove_tmp_files()
    _remove_tmp_files(r"[a-zA-Z_]*_[\d]+_[\d]+.json")


@pytest.fixture
def json_file_good_data_op_fan_mode() -> Generator:
    """Return json test file with good structure.

    Yields:
        json's path file
    """
    yield Path.cwd().joinpath("tests").joinpath("data").joinpath("good_data_op_fan_mode.json")
    _remove_tmp_files()
    _remove_tmp_files(r"[a-zA-Z_]*_[\d]+_[\d]+.json")


@pytest.fixture
def json_file_good_data_op_fan_swing_mode() -> Generator:
    """Return json test file with good structure.

    Yields:
        json's path file
    """
    yield Path.cwd().joinpath("tests").joinpath("data").joinpath("good_data_op_fan_swing_mode.json")
    _remove_tmp_files()
    _remove_tmp_files(r"[a-zA-Z_]*_[\d]+_[\d]+.json")


@pytest.fixture
def json_file_good_data_op_swing_mode() -> Generator:
    """Return json test file with good structure.

    Yields:
        json's path file
    """
    yield Path.cwd().joinpath("tests").joinpath("data").joinpath("good_data_op_swing_mode.json")
    _remove_tmp_files()
    _remove_tmp_files(r"[a-zA-Z_]*_[\d]+_[\d]+.json")


@pytest.fixture
def json_file_partial_dict_op_fan_swing_mode() -> Path:
    """Return json test file with good structure for partial load.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("partial_dicts").joinpath("good_data_op_fan_swing_mode.json")


@pytest.fixture
def json_file_previous_partial_dict_op_fan_swing_mode() -> Path:
    """Return json test file with good structure for partial load.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("partial_dicts").joinpath("good_data_op_fan_swing_mode_tmp_003.json")


@pytest.fixture
def json_file_partial_dict_op_swing_mode() -> Path:
    """Return json test file with good structure for partial load.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("partial_dicts").joinpath("good_data_op_swing_mode.json")


@pytest.fixture
def json_file_last_previous_partial_dict_op_swing_mode() -> Path:
    """Return json test file with good structure for partial load.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("partial_dicts").joinpath("good_data_op_swing_mode_tmp_003.json")


@pytest.fixture
def json_file_missing_operation_modes() -> Path:
    """Return json test file without operation modes.

    Returns:
        json's path file
    """
    return Path.cwd().joinpath("tests").joinpath("data").joinpath("missing_required_operation_modes.json")


@pytest.fixture(scope="function")
def runner(request):
    """Pytest runner fixture.

    Arguments:
        request: pytest request

    Returns:
        Click CliRunner object
    """
    return CliRunner()


def dict_from_json(json_file: Path) -> dict:
    """Obtain json content from file path.

    Arguments:
        json_file: path of json file

    Returns:
        json's content as dict
    """
    with open(str(json_file), "r", encoding='utf-8') as in_file:
        _expected_dict = json.load(in_file)
    return _expected_dict


@dataclass
class ExpectedValues:
    """Test expected values and bytes that generate them."""

    code_inc: bytes
    code_dec: bytes
    code_even: bytes
    code_odd: bytes
    code_upper: bytes
    code_lower: bytes
    code_last: bytes
    code_last_lower: bytes
    expected_inc: str
    expected_dec: str
    expected_even: str
    expected_odd: str
    expected_upper: str
    expected_lower: str
    expected_last: str
    expected_last_lower: str

    @staticmethod
    def _bytes_to_str(code: bytes) -> str:
        b64_data = binascii.b2a_base64(code, newline=False)
        return b64_data.decode('utf-8')

    def __init__(self):
        """Object initialization."""
        self.code_inc = b'12345678'
        self.code_dec = b'87654321'
        self.code_even = b'024681012'
        self.code_odd = b'135791113'
        self.code_upper = b'ABCDEFG'
        self.code_lower = b'abcdefg'
        self.code_last = b'TUVWXYZ'
        self.code_last_lower = b'tuvwxyz'

        self.expected_inc = ExpectedValues._bytes_to_str(self.code_inc)
        self.expected_dec = ExpectedValues._bytes_to_str(self.code_dec)
        self.expected_even = ExpectedValues._bytes_to_str(self.code_even)
        self.expected_odd = ExpectedValues._bytes_to_str(self.code_odd)
        self.expected_upper = ExpectedValues._bytes_to_str(self.code_upper)
        self.expected_lower = ExpectedValues._bytes_to_str(self.code_lower)
        self.expected_last = ExpectedValues._bytes_to_str(self.code_last)
        self.expected_last_lower = ExpectedValues._bytes_to_str(self.code_last_lower)



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Test Broadlink manager module."""
import binascii
from unittest.mock import Mock, patch

import broadlink.remote
import pytest
from broadlink import gendevice
from broadlink.const import DEFAULT_PORT
from broadlink.exceptions import ReadError

from broadlink_listener.cli_tools.broadlink_manager import BroadlinkManager


class TestBroadlinkManager:
    """Broadlink manager test class."""

    @patch('broadlink.device.Device.auth', Mock(return_value=True))
    @patch('broadlink.gendevice')
    def test_init(self, patched_auth):
        """Test initialization.

        Arguments:
            patched_auth: patched version of broadlink Device's auth method.
        """
        _ = BroadlinkManager('0x1234', '192.168.1.1', '12345678')

        with pytest.raises(ValueError):
            _ = BroadlinkManager('0x1234', '192.168.1.1', '1234567ssss8')

        with pytest.raises(ValueError):
            _ = BroadlinkManager('abcs', '192.168.1.1', '12345678')

    def test_rmmini_device(self):
        """Test rmmini device creation with gendevice."""
        _a = gendevice(0x51DA, ('192.168.1.1', DEFAULT_PORT), '12345678')
        assert issubclass(type(_a), broadlink.remote.rmmini)
        assert hasattr(_a, 'enter_learning')

    @patch('broadlink.device.Device.auth', Mock(return_value=True))
    def test_broadlink_manager_device(self):
        """Test rmmini device creation with BroadlinkManager."""
        _a = BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
        assert issubclass(type(_a.device), broadlink.remote.rmmini)
        assert hasattr(_a.device, 'enter_learning')

    def test_learn_code(self):
        """Test rmmini device learn mode."""
        _code_value = b'12345678'
        b64_data = binascii.b2a_base64(_code_value, newline=False)
        _expected = b64_data.decode('utf-8')

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('broadlink.remote.rmmini.check_data', Mock(return_value=_code_value)):
            _a = BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            _code = _a.learn_single_code()
            assert _expected == _code

    def test_learn_code_exception(self):
        """Test rmmini device learn mode, exception case will retrieve None ."""
        _expected = None

        with patch('time.sleep'), patch('broadlink_listener.cli_tools.broadlink_manager.DEFAULT_TIMEOUT', 1), patch(
            'broadlink.remote.rmmini.enter_learning'
        ), patch('broadlink.device.Device.auth', Mock(return_value=True)), patch(
            'broadlink.remote.rmmini.check_data', Mock(side_effect=ReadError(-10))
        ):

            _a = BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            _code = _a.learn_single_code()

            assert _expected == _code



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Test utils module."""

import logging
import socket
from unittest.mock import Mock, patch

from broadlink_listener.cli_tools.utils import configure_logger, get_local_ip_address


@patch('logging.basicConfig')
def test_logger(patched_log):
    """Testing configure_logger.

    Arguments:
        patched_log: patched basicConfig method
    """
    configure_logger()
    patched_log.assert_called_with(
        level=logging.INFO, format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s', handlers=None
    )

    configure_logger('info')
    patched_log.assert_called_with(
        level=logging.INFO, format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s', handlers=None
    )

    configure_logger('debug')
    patched_log.assert_called_with(
        level=logging.DEBUG,
        format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s',
        handlers=None,
    )

    configure_logger('warning')
    patched_log.assert_called_with(
        level=logging.WARN, format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s', handlers=None
    )

    configure_logger('error')
    patched_log.assert_called_with(
        level=logging.ERROR,
        format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s',
        handlers=None,
    )


@patch('socket.socket.getsockname', Mock(side_effect=TimeoutError()))
def test_get_ip_localhost():
    """Testing get_local_ip_address with localhost address."""
    assert get_local_ip_address() == "127.0.0.1"


@patch('socket.socket.getsockname', Mock(return_value=['192.168.1.150']))
def test_get_ip_real_ip_patched():
    """Testing get_local_ip_address with real mocked address."""
    assert get_local_ip_address() == '192.168.1.150'


def test_get_ip_real_ip():
    """Testing get_local_ip_address with real mocked address."""
    with patch('socket.socket.connect'), patch('socket.socket'), patch(
        'socket.socket.getsockname', Mock(return_value='192.168.1.150')
    ):
        assert socket.socket.getsockname() == '192.168.1.150'



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Broadlink device manager."""

import binascii
import time
from typing import Optional, cast

import click
from broadlink import gendevice
from broadlink.const import DEFAULT_PORT, DEFAULT_TIMEOUT
from broadlink.exceptions import ReadError, StorageError
from broadlink.remote import rmmini


class BroadlinkManager:  # pylint: disable=too-few-public-methods
    """Manager class for Broadlink device."""

    def __init__(self, dev_type: str, ip_addr: str, mac_addr: str):
        """Broadlink Manager.

        Arguments:
            dev_type: Broadlink device type from discover IR command
            ip_addr: Broadlink IP address from discover IR command
            mac_addr: Broadlink MAC address from discover IR command
        """
        self.__dev: rmmini = cast(rmmini, gendevice(int(dev_type, 0), (ip_addr, DEFAULT_PORT), mac_addr))
        self.__dev.auth()

    @property
    def device(self) -> rmmini:
        """Broadlink device.

        Returns:
            rmmini device object
        """
        return self.__dev

    def learn_single_code(self) -> Optional[str]:
        """Process to learn single IR code.

        Returns:
            Optional[str]: str if IR code was listened, None otherwise
        """
        click.echo("Listening...")
        self.device.enter_learning()
        start = time.time()
        _ret = None
        while time.time() - start < DEFAULT_TIMEOUT:
            time.sleep(1)
            try:
                data: bytes = self.device.check_data()
            except (ReadError, StorageError):
                continue
            else:
                b64_data = binascii.b2a_base64(data, newline=False)
                _ret = b64_data.decode('utf-8')
                break

        return _ret



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Module with utility methods."""

import logging
import socket


def configure_logger(loglevel: str = 'info') -> None:
    """Configure logger facility, from clients or server side.

    Args:
        loglevel: level of logging facility
    """
    # configure logging
    _log_level = logging.INFO
    if loglevel.lower() == 'debug':
        _log_level = logging.DEBUG
    elif loglevel.lower() == 'warning':
        _log_level = logging.WARNING
    elif loglevel.lower() == 'error':
        _log_level = logging.ERROR

    logging.basicConfig(
        level=_log_level,
        format='%(asctime)s [%(levelname)s - %(filename)s:%(lineno)d]    %(message)s',
        handlers=None,
    )


def get_local_ip_address() -> str:
    """Returns local IP address for main interface.

    Returns:
        the local host's IP address configured
    """
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as _sock:
        _sock.settimeout(0)
        try:
            # doesn't even have to be reachable
            _sock.connect(('10.254.254.254', 1))
            _ip = _sock.getsockname()[0]
        except (TimeoutError, InterruptedError, Exception):  # pylint: disable=broad-except
            _ip = '127.0.0.1'
    return _ip



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Tests for `broadlink_listener` package."""

from unittest.mock import Mock, patch

from broadlink_listener import __version__
from broadlink_listener.cli_tools import cli


def test_command_line_interface_help(runner):
    """Test the CLI.

    Arguments:
        runner: pytest click CliRunner object
    """
    help_result = runner.invoke(cli.main, ['--help'])
    assert help_result.exit_code == 0
    assert '  --help     Show this message and exit.' in help_result.output


def test_command_line_interface_version(runner):
    """Test the CLI.

    Arguments:
        runner: pytest click CliRunner object
    """
    help_result = runner.invoke(cli.main, ['--version'])
    assert help_result.exit_code == 0
    assert __version__ in help_result.output


def test_generation_smart_ir(runner, json_file_good_data_op_swing_mode):
    """Test the CLI.

    Arguments:
        runner: pytest click CliRunner object
        json_file_good_data_op_swing_mode: existing json file
    """
    with patch('broadlink.remote.rmmini.enter_learning'), patch(
        'broadlink.device.Device.auth', Mock(return_value=True)
    ), patch('broadlink_listener.cli_tools.broadlink_manager.BroadlinkManager'), patch(
        'broadlink_listener.cli_tools.smartir_manager.SmartIrManager.learn_off'
    ) as _learn_off, patch(
        'broadlink_listener.cli_tools.smartir_manager.SmartIrManager.learn_all'
    ) as _learn_all, patch(
        'broadlink_listener.cli_tools.smartir_manager.SmartIrManager.save_dict'
    ) as _save:
        _ = runner.invoke(
            cli.main, ['generate-smart-ir', str(json_file_good_data_op_swing_mode), '0x1234', '192.168.1.1', '12345678']
        )
        _learn_off.assert_called_once()
        _learn_all.assert_called_once()
        _save.assert_called_once()


def test_py_version():
    """Dummy test to print python version used by pytest."""
    import sys

    print(f"in TEST: {sys.version}  -- {sys.version_info}")
    # if sys.version_info <= (3, 9, 18):
    #     # 3.9 OK
    #     assert True
    # else:
    #     # 3.10 FAIL
    #     assert False



# SPDX-FileCopyrightText: 2022 Gabriele Pongelli
#
# SPDX-License-Identifier: MIT

"""Test smartir manager module."""

from itertools import cycle
from unittest.mock import Mock, patch

import click
import pytest
from freezegun import freeze_time

from broadlink_listener.cli_tools.smartir_manager import BroadlinkManager, SmartIrManager
from tests.conftest import ExpectedValues, dict_from_json


class TestSmartIR:
    """SmartIR manager test class."""

    def test_not_broadlink(self, json_file_not_broadlink):
        """Test controller not supported.

        Arguments:
            json_file_not_broadlink: pytest fixture
        """
        with pytest.raises(click.exceptions.UsageError):
            _ = SmartIrManager(json_file_not_broadlink, Mock())

    def test_not_base64(self, json_file_not_base64):
        """Test encoding not supported.

        Arguments:
            json_file_not_base64: pytest fixture
        """
        with pytest.raises(click.exceptions.UsageError):
            _ = SmartIrManager(json_file_not_base64, Mock())

    def test_missing_min_temp(self, json_file_missing_min_temp):
        """Test missing min temp.

        Arguments:
            json_file_missing_min_temp: pytest fixture
        """
        with pytest.raises(click.exceptions.UsageError):
            _ = SmartIrManager(json_file_missing_min_temp, Mock())

    def test_missing_max_temp(self, json_file_missing_max_temp):
        """Test missing max temp.

        Arguments:
            json_file_missing_max_temp: pytest fixture
        """
        with pytest.raises(click.exceptions.UsageError):
            _ = SmartIrManager(json_file_missing_max_temp, Mock())

    def test_missing_operation_modes(self, json_file_missing_operation_modes):
        """Test missing operation mode.

        Arguments:
            json_file_missing_operation_modes: pytest fixture
        """
        with pytest.raises(click.exceptions.UsageError):
            _ = SmartIrManager(json_file_missing_operation_modes, Mock())

    def test_learning_op_mode(self, json_file_good_data_op_mode):
        """Test dict generation, operationMode only.

        Arguments:
            json_file_good_data_op_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_mode)

        _expect_dict_before_learn = dict(_expected_dict)
        _expect_dict_before_learn['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                '18': '',
                '19': '',
                '20': '',
            },
            'heat': {
                '18': '',
                '19': '',
                '20': '',
            },
        }

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                '18': _expected_values.expected_inc,
                '19': _expected_values.expected_dec,
                '20': _expected_values.expected_odd,
            },
            'heat': {
                '18': _expected_values.expected_even,
                '19': _expected_values.expected_lower,
                '20': _expected_values.expected_upper,
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(json_file_good_data_op_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678'))
            assert _expect_dict_before_learn == _a.smartir_dict

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_learning_op_fan_mode(self, json_file_good_data_op_fan_mode):
        """Test dict generation, operationMode and fanMode only.

        Arguments:
            json_file_good_data_op_fan_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_fan_mode)

        _expect_dict_before_learn = dict(_expected_dict)
        _expect_dict_before_learn['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
                'high': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
            },
            'heat': {
                'low': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
                'high': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
            },
        }

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    '18': _expected_values.expected_inc,
                    '19': _expected_values.expected_dec,
                    '20': _expected_values.expected_odd,
                },
                'high': {
                    '18': _expected_values.expected_even,
                    '19': _expected_values.expected_lower,
                    '20': _expected_values.expected_upper,
                },
            },
            'heat': {
                'low': {
                    '18': _expected_values.expected_last,
                    '19': _expected_values.expected_last_lower,
                    '20': _expected_values.expected_inc,
                },
                'high': {
                    '18': _expected_values.expected_dec,
                    '19': _expected_values.expected_odd,
                    '20': _expected_values.expected_even,
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(json_file_good_data_op_fan_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678'))
            assert _expect_dict_before_learn == _a.smartir_dict

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_learning_op_swing_mode(self, json_file_good_data_op_swing_mode):
        """Test dict generation, operationMode and swingMode only.

        Arguments:
            json_file_good_data_op_swing_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_swing_mode)

        _expect_dict_before_learn = dict(_expected_dict)
        _expect_dict_before_learn['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'up': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
                'down': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
            },
            'heat': {
                'up': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
                'down': {
                    '18': '',
                    '19': '',
                    '20': '',
                },
            },
        }

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'up': {
                    '18': _expected_values.expected_inc,
                    '19': _expected_values.expected_dec,
                    '20': _expected_values.expected_odd,
                },
                'down': {
                    '18': _expected_values.expected_even,
                    '19': _expected_values.expected_lower,
                    '20': _expected_values.expected_upper,
                },
            },
            'heat': {
                'up': {
                    '18': _expected_values.expected_last,
                    '19': _expected_values.expected_last_lower,
                    '20': _expected_values.expected_inc,
                },
                'down': {
                    '18': _expected_values.expected_dec,
                    '19': _expected_values.expected_odd,
                    '20': _expected_values.expected_even,
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_good_data_op_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            )
            assert _expect_dict_before_learn == _a.smartir_dict

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_learning_op_fan_swing_mode(self, json_file_good_data_op_fan_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_fan_swing_mode)

        _expect_dict_before_learn = dict(_expected_dict)
        _expect_dict_before_learn['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    'up': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                    'down': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                },
                'high': {
                    'up': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                    'down': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                },
            },
            'heat': {
                'low': {
                    'up': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                    'down': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                },
                'high': {
                    'up': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                    'down': {
                        '18': '',
                        '19': '',
                        '20': '',
                    },
                },
            },
        }

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_inc,
                        '19': _expected_values.expected_dec,
                        '20': _expected_values.expected_odd,
                    },
                    'down': {
                        '18': _expected_values.expected_even,
                        '19': _expected_values.expected_lower,
                        '20': _expected_values.expected_upper,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_last,
                        '19': _expected_values.expected_last_lower,
                        '20': _expected_values.expected_inc,
                    },
                    'down': {
                        '18': _expected_values.expected_dec,
                        '19': _expected_values.expected_odd,
                        '20': _expected_values.expected_even,
                    },
                },
            },
            'heat': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_lower,
                        '19': _expected_values.expected_upper,
                        '20': _expected_values.expected_last,
                    },
                    'down': {
                        '18': _expected_values.expected_last_lower,
                        '19': _expected_values.expected_inc,
                        '20': _expected_values.expected_dec,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_odd,
                        '19': _expected_values.expected_even,
                        '20': _expected_values.expected_lower,
                    },
                    'down': {
                        '18': _expected_values.expected_upper,
                        '19': _expected_values.expected_last,
                        '20': _expected_values.expected_last_lower,
                    },
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_good_data_op_fan_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            )
            assert _expect_dict_before_learn == _a.smartir_dict

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_partial_op_fan_swing_mode(
        self, json_file_partial_dict_op_fan_swing_mode, json_file_previous_partial_dict_op_fan_swing_mode
    ):
        """Test dict generation, all fields.

        Arguments:
            json_file_partial_dict_op_fan_swing_mode: json file
            json_file_previous_partial_dict_op_fan_swing_mode: json file with partial IR saved
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_previous_partial_dict_op_fan_swing_mode)
        _source_dict = dict_from_json(json_file_partial_dict_op_fan_swing_mode)

        _expected_dict.update({'off': _source_dict['commands']['off']})

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_partial_dict_op_fan_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            )
            assert _a.smartir_dict['commands'] == _expected_dict
            assert _a.partial_inc == 3

    def test_partial_op_swing_mode_multiple_files(
        self, json_file_partial_dict_op_swing_mode, json_file_last_previous_partial_dict_op_swing_mode
    ):
        """Test dict generation, all fields.

        Arguments:
            json_file_partial_dict_op_swing_mode: json file
            json_file_last_previous_partial_dict_op_swing_mode: last json file with partial IR saved
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_last_previous_partial_dict_op_swing_mode)
        _source_dict = dict_from_json(json_file_partial_dict_op_swing_mode)

        _expected_dict.update({'off': _source_dict['commands']['off']})

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_partial_dict_op_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            )
            assert _a.smartir_dict['commands'] == _expected_dict
            assert _a.partial_inc == 3

    def test_skip_temp(self, json_file_good_data_op_fan_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_fan_swing_mode)

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_inc,
                        '19': _expected_values.expected_dec,
                        '20': _expected_values.expected_odd,
                    },
                    'down': {
                        '18': _expected_values.expected_even,
                        '19': _expected_values.expected_lower,
                        '20': _expected_values.expected_upper,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_last,
                        '19': _expected_values.expected_last_lower,
                        '20': _expected_values.expected_inc,
                    },
                    'down': {
                        '18': _expected_values.expected_dec,
                        '19': _expected_values.expected_odd,
                        '20': _expected_values.expected_even,
                    },
                },
            },
            'heat': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_lower,
                        '19': _expected_values.expected_lower,
                        '20': _expected_values.expected_lower,
                    },
                    'down': {
                        '18': _expected_values.expected_upper,
                        '19': _expected_values.expected_upper,
                        '20': _expected_values.expected_upper,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_last,
                        '19': _expected_values.expected_last,
                        '20': _expected_values.expected_last,
                    },
                    'down': {
                        '18': _expected_values.expected_last_lower,
                        '19': _expected_values.expected_last_lower,
                        '20': _expected_values.expected_last_lower,
                    },
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_good_data_op_fan_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678'), ('heat',)
            )

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_skip_swing(self, json_file_good_data_op_fan_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_fan_swing_mode)

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_inc,
                        '19': _expected_values.expected_dec,
                        '20': _expected_values.expected_odd,
                    },
                    'down': {
                        '18': _expected_values.expected_even,
                        '19': _expected_values.expected_lower,
                        '20': _expected_values.expected_upper,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_last,
                        '19': _expected_values.expected_last_lower,
                        '20': _expected_values.expected_inc,
                    },
                    'down': {
                        '18': _expected_values.expected_dec,
                        '19': _expected_values.expected_odd,
                        '20': _expected_values.expected_even,
                    },
                },
            },
            'heat': {
                'low': {
                    'up': {
                        '18': _expected_values.expected_lower,
                        '19': _expected_values.expected_upper,
                        '20': _expected_values.expected_last,
                    },
                    'down': {
                        '18': _expected_values.expected_lower,
                        '19': _expected_values.expected_upper,
                        '20': _expected_values.expected_last,
                    },
                },
                'high': {
                    'up': {
                        '18': _expected_values.expected_last_lower,
                        '19': _expected_values.expected_inc,
                        '20': _expected_values.expected_dec,
                    },
                    'down': {
                        '18': _expected_values.expected_last_lower,
                        '19': _expected_values.expected_inc,
                        '20': _expected_values.expected_dec,
                    },
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_good_data_op_fan_swing_mode,
                BroadlinkManager('0x51DA', '192.168.1.1', '12345678'),
                (),
                ('heat',),
            )

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_skip_swing_no_fan_mode(self, json_file_good_data_op_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_swing_mode: json file
        """
        _expected_values = ExpectedValues()
        _expected_dict = dict_from_json(json_file_good_data_op_swing_mode)

        _expected_dict['commands'] = {
            'off': _expected_dict['commands']['off'],
            'cool': {
                'up': {
                    '18': _expected_values.expected_inc,
                    '19': _expected_values.expected_dec,
                    '20': _expected_values.expected_odd,
                },
                'down': {
                    '18': _expected_values.expected_even,
                    '19': _expected_values.expected_lower,
                    '20': _expected_values.expected_upper,
                },
            },
            'heat': {
                'up': {
                    '18': _expected_values.expected_last,
                    '19': _expected_values.expected_last_lower,
                    '20': _expected_values.expected_inc,
                },
                'down': {
                    '18': _expected_values.expected_last,
                    '19': _expected_values.expected_last_lower,
                    '20': _expected_values.expected_inc,
                },
            },
        }

        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink.remote.rmmini.check_data',
            Mock(
                side_effect=cycle(
                    [
                        _expected_values.code_inc,
                        _expected_values.code_dec,
                        _expected_values.code_odd,
                        _expected_values.code_even,
                        _expected_values.code_lower,
                        _expected_values.code_upper,
                        _expected_values.code_last,
                        _expected_values.code_last_lower,
                    ]
                )
            ),
        ):
            _a = SmartIrManager(
                json_file_good_data_op_swing_mode,
                BroadlinkManager('0x51DA', '192.168.1.1', '12345678'),
                (),
                ('heat',),
            )

            _a.learn_all()
            assert _expected_dict == _a.smartir_dict

    def test_skip_swing_cli_param(self, json_file_good_data_op_fan_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_mode: json file without swing
        """
        with patch('broadlink.device.Device.auth', Mock(return_value=True)):
            with pytest.raises(click.exceptions.UsageError):
                _ = SmartIrManager(
                    json_file_good_data_op_fan_mode,
                    BroadlinkManager('0x51DA', '192.168.1.1', '12345678'),
                    (),
                    ('heat',),
                )

    def test_mode_not_present_from_swing(self, json_file_good_data_op_fan_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
        """
        with patch('broadlink.device.Device.auth', Mock(return_value=True)):
            with pytest.raises(click.exceptions.UsageError):
                _ = SmartIrManager(
                    json_file_good_data_op_fan_swing_mode,
                    BroadlinkManager('0x51DA', '192.168.1.1', '12345678'),
                    (),
                    ('hesat',),
                )

    def test_mode_not_present_from_temp(self, json_file_good_data_op_fan_swing_mode):
        """Test dict generation, all fields.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
        """
        with patch('broadlink.device.Device.auth', Mock(return_value=True)):
            with pytest.raises(click.exceptions.UsageError):
                _ = SmartIrManager(
                    json_file_good_data_op_fan_swing_mode,
                    BroadlinkManager('0x51DA', '192.168.1.1', '12345678'),
                    ('hesat',),
                    (),
                )

    def test_no_code(self, json_file_good_data_op_mode):
        """Test dict generation, operationMode only.

        Arguments:
            json_file_good_data_op_mode: json file
        """
        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'), patch(
            'broadlink_listener.cli_tools.broadlink_manager.BroadlinkManager.learn_single_code', Mock(return_value=None)
        ):
            with pytest.raises(click.exceptions.UsageError):
                _a = SmartIrManager(json_file_good_data_op_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678'))
                _a.learn_all()

            with pytest.raises(click.exceptions.UsageError):
                _a = SmartIrManager(json_file_good_data_op_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678'))
                _a.learn_off()

    @freeze_time("2023-02-10 12:10:30")
    def test_handle_signal(self, json_file_good_data_op_fan_swing_mode, capsys):
        """Test handle keyboard interrupt.

        Arguments:
            json_file_good_data_op_fan_swing_mode: json file
            capsys: pytest mock to capture stdout
        """
        with patch('broadlink.remote.rmmini.enter_learning'), patch(
            'broadlink.device.Device.auth', Mock(return_value=True)
        ), patch('time.sleep'), patch('builtins.input'):
            _a = SmartIrManager(
                json_file_good_data_op_fan_swing_mode, BroadlinkManager('0x51DA', '192.168.1.1', '12345678')
            )
            _a.save_dict()

            captured = capsys.readouterr()
            assert 'good_data_op_fan_swing_mode_20230210_121030.json' in captured.out



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['broadlink_listener', 'broadlink_listener.cli_tools', 'tests']

package_data = \
{'': ['*'], 'tests': ['data/*', 'partial_dicts/*']}

install_requires = \
['broadlink>=0.18.3,<0.19.0',
 'click==8.1.3',
 'cloup>=2.0.0,<3.0.0',
 'termcolor>=2.1.1,<3.0.0',
 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['broadlink-listener = '
                     'broadlink_listener.cli_tools.cli:main']}

setup_kwargs = {
    'name': 'broadlink-listener',
    'version': '1.2.0',
    'description': 'Broadlink IR codes listener and SmartIR json generator.',
    'long_description': '# Broadlink Listener\n\n\n[![pypi](https://img.shields.io/pypi/v/broadlink-listener.svg)](https://pypi.org/project/broadlink-listener/)\n[![python](https://img.shields.io/pypi/pyversions/broadlink-listener.svg)](https://pypi.org/project/broadlink-listener/)\n[![Build Status](https://github.com/gpongelli/broadlink-listener/actions/workflows/dev.yml/badge.svg)](https://github.com/gpongelli/broadlink-listener/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/gpongelli/broadlink-listener/branch/main/graphs/badge.svg)](https://codecov.io/github/gpongelli/broadlink-listener)\n\n\n\nBroadlink IR codes listener and SmartIR json generator.\n\nThis project will install a `broadlink-listener` command line tool that can be used to generate a climate [SmartIR](https://github.com/smartHomeHub/SmartIR)\ncompatible json, starting from an initial structure that defines climate behavior, putting Broadlink IR remote to\nlistening state, until all IR code combination will being scan.\n\n\n* Documentation: <https://gpongelli.github.io/broadlink-listener>\n* GitHub: <https://github.com/gpongelli/broadlink-listener>\n* PyPI: <https://pypi.org/project/broadlink-listener/>\n* Free software: MIT\n\n\n## Features\n\n* Discover Broadlink IR remote\n* Starting from SmartIR json structure like\n```json\n{\n  "supportedController": "Broadlink",\n  "minTemperature": 16,\n  "maxTemperature": 31,\n  "precision": 1,\n  "operationModes": [\n    "op_a",\n    "op_b"\n  ],\n  "fanModes": [\n    "fan_a",\n    "fan_b"\n  ],\n  "swingModes": [\n    "swing_a",\n    "swing_b"\n  ]\n}\n```\n  it helps you listen all the defined IR codes to create a json like\n```json\n{\n  "supportedController": "Broadlink",\n  "minTemperature": 16,\n  "maxTemperature": 31,\n  "precision": 1,\n  "operationModes": [\n    "op_a",\n    "op_b"\n  ],\n  "fanModes": [\n    "fan_a",\n    "fan_b"\n  ],\n  "swingModes": [\n    "swing_a",\n    "swing_b"\n  ],\n  "commands": {\n    "off": "...",\n    "op_a": {\n        "fan_a": {\n            "swing_a": {\n                "16": "....",\n\n                "31": "...."\n            },\n            "swing_b": {\n                "16": "....",\n\n                "31": "...."\n            }\n        },\n        "fan_b": {\n            "swing_a": {\n                "16": "....",\n\n                "31": "...."\n            },\n            "swing_b": {\n                "16": "....",\n\n                "31": "...."\n            }\n        }\n    },\n    "op_b": {\n        "fan_a": {\n            "swing_a": {\n                "16": "....",\n\n                "31": "...."\n            },\n            "swing_b": {\n                "16": "....",\n\n                "31": "...."\n            }\n        },\n        "fan_b": {\n            "swing_a": {\n                "16": "....",\n\n                "31": "...."\n            },\n            "swing_b": {\n                "16": "....",\n\n                "31": "...."\n            }\n        }\n    }\n  }\n}\n```\n\n* Mandatory fields into starting json\n  * `supportedController`, `minTemperature`, `maxTemperature`, `precision`\n* Optional fields (at least one must be present or nothing will be listened):\n  * `operationModes`, `fanModes`,`swingModes`\n* Generated file can be used into SmartIR HomeAssistant component\n* It\'s possible to interrupt with CTRL-C at any moment, a temporary file will be saved\n* Temporary files are also saved at the end of each temperature range\n* In case of existing temporary file, the already learnt combination will be skipped\n\n\n## Example\n\nExample of cli command:\n```bash\n$ broadlink-listener generate-smart-ir ./real_data/1124.json <DEVICE_TYPE> <IP> <MAC_ADDR> -n dry -n fan_only -s eco_cool\n```\n\n`real_data/1124.json` file is [this one from SmartIR GitHub repo](https://github.com/smartHomeHub/SmartIR/blob/master/codes/climate/1124.json)\nin which I\'ve added the missing "swingModes" array, supported by climate but not present on json:\n```json\n"swingModes": [\n  "auto",\n  "high",\n  "mid_high",\n  "middle",\n  "mid_low",\n  "low",\n  "swing"\n],\n```\n\n`<DEVICE_TYPE>`, `<IP>`, `<MAC_ADDR>` parameter can be obtained running:\n```bash\n$ broadlink-listener discover_ir\n```\n\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter)\nand the [gpongelli/cookiecutter-pypackage](https://github.com/gpongelli/cookiecutter-pypackage) project template.\n',
    'author': 'Gabriele Pongelli',
    'author_email': 'gabriele.pongelli@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gpongelli/broadlink-listener',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<4.0',
}


setup(**setup_kwargs)
